package com.mortennobel.mandelbrot;

import com.sun.opengl.util.Animator;

import java.awt.*;
import java.io.*;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.glu.GLU;

public class MandelbrotJOGL extends GLCanvas implements GLEventListener, MandelbrotRender {
	private boolean updateUniformVars = true;
	private int vertexShaderProgram;
	private int fragmentShaderProgram;
	private int shaderprogram;
	private final MandelbrotSetting settings;
	private final Animator animator = new Animator(this);

	private FPSStat fpsStat = new FPSStat();
	
	public MandelbrotJOGL(MandelbrotSetting settings) {
		this.settings = settings;
		animator.setRunAsFastAsPossible(true);
		addGLEventListener(this);
    }

    public void init(GLAutoDrawable drawable) {
        GL gl = drawable.getGL();

        // Enable VSync
        gl.setSwapInterval(1);
		gl.glShadeModel(GL.GL_FLAT);
		try {
			attachShaders(gl);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private String[] loadShaderSrc(String name){
		StringBuilder sb = new StringBuilder();
		try{
			InputStream is = getClass().getResourceAsStream(name);
			BufferedReader br = new BufferedReader(new InputStreamReader(is));
			String line;
			while ((line = br.readLine())!=null){
				sb.append(line);
				sb.append('\n');
			}
			is.close();
		}
		catch (Exception e){
			e.printStackTrace();
		}
		return new String[]{sb.toString()};
	}

	private void attachShaders(GL gl) throws Exception {
		vertexShaderProgram = gl.glCreateShader(GL.GL_VERTEX_SHADER);
		fragmentShaderProgram = gl.glCreateShader(GL.GL_FRAGMENT_SHADER);

		String[] vsrc = loadShaderSrc("mandelbrot.vs");
		gl.glShaderSource(vertexShaderProgram, 1, vsrc, null, 0);
		gl.glCompileShader(vertexShaderProgram);

		String[] fsrc = loadShaderSrc("mandelbrot.fs");
		gl.glShaderSource(fragmentShaderProgram, 1, fsrc, null, 0);
		gl.glCompileShader(fragmentShaderProgram);

		shaderprogram = gl.glCreateProgram();
		gl.glAttachShader(shaderprogram, vertexShaderProgram);
		gl.glAttachShader(shaderprogram, fragmentShaderProgram);
		gl.glLinkProgram(shaderprogram);
		gl.glValidateProgram(shaderprogram);
		IntBuffer intBuffer = IntBuffer.allocate(1);
		gl.glGetProgramiv(shaderprogram, GL.GL_LINK_STATUS,intBuffer);
		if (intBuffer.get(0)!=1){
			gl.glGetProgramiv(shaderprogram, GL.GL_INFO_LOG_LENGTH,intBuffer);
			int size = intBuffer.get(0);
			System.err.println("Program link error: ");
			if (size>0){
				ByteBuffer byteBuffer = ByteBuffer.allocate(size);
				gl.glGetProgramInfoLog(shaderprogram, size, intBuffer, byteBuffer);
				for (byte b:byteBuffer.array()){
					System.err.print((char)b);
				}
			} else {
				System.out.println("Unknown error");
			}
			System.exit(1);
		}
		gl.glUseProgram(shaderprogram);
	}

	public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
        GL gl = drawable.getGL();
        GLU glu = new GLU();

        if (height <= 0) { // avoid a divide by zero error!

            height = 1;
        }
        gl.glViewport(0, 0, width, height);
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        glu.gluOrtho2D(0, 1, 0, 1);
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glLoadIdentity();
    }

    public void display(GLAutoDrawable drawable) {
		fpsStat.timerStart();
		GL gl = drawable.getGL();

		if (updateUniformVars){
			updateUniformVars(gl);
		}

		// Reset the current matrix to the "identity"
		gl.glLoadIdentity();

        // Draw A Quad
        gl.glBegin(GL.GL_QUADS);
		{
			gl.glTexCoord2f(0.0f, 0.0f);
			gl.glVertex3f(0.0f, 1.0f, 1.0f);
			gl.glTexCoord2f(1.0f, 0.0f);
			gl.glVertex3f(1.0f, 1.0f, 1.0f);
			gl.glTexCoord2f(1.0f, 1.0f);
			gl.glVertex3f(1.0f, 0.0f, 1.0f);
			gl.glTexCoord2f(0.0f, 1.0f);
			gl.glVertex3f(0.0f, 0.0f, 1.0f);
		}
		// Done Drawing The Quad
        gl.glEnd();

        // Flush all drawing operations to the graphics card
        gl.glFlush();
		fpsStat.timerEnd();
    }

	private void updateUniformVars(GL gl) {
        // get memory address of uniform shader variables
		int mandel_x = gl.glGetUniformLocation(shaderprogram, "mandel_x");
		int mandel_y = gl.glGetUniformLocation(shaderprogram, "mandel_y");
		int mandel_width = gl.glGetUniformLocation(shaderprogram, "mandel_width");
		int mandel_height = gl.glGetUniformLocation(shaderprogram, "mandel_height");
		int mandel_iterations = gl.glGetUniformLocation(shaderprogram, "mandel_iterations");
		assert(mandel_x!=-1);
		assert(mandel_y!=-1);
		assert(mandel_width!=-1);
		assert(mandel_height!=-1);
		assert(mandel_iterations!=-1);
        // set uniform shader variables
		gl.glUniform1f(mandel_x, settings.getX());
		gl.glUniform1f(mandel_y, settings.getY());
		gl.glUniform1f(mandel_width, settings.getWidth());
		gl.glUniform1f(mandel_height, settings.getHeight());
		gl.glUniform1f(mandel_iterations, settings.getIterations());

	}

	public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) {
    }

	public void pauseAnimation() {
		animator.stop();
	}

	public void resumeAnimation() {
		animator.start();
	}

	public Component getComponent() {
		return this;
	}

	public FPSStat getStats() {
		return fpsStat;
	}
}


